#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <pwd.h>
#include <ctype.h>
#include <unistd.h>

#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include "misc.h"

/****************************************************************************/
/* Does       : check if a local port is connected                          */
/* Parameters : p - port number in host format                              */
/* Returns    : greater than zero if connected, zero if not, negative on    */
/*              failure                                                     */
/* Notes      : could be done in /proc/net/tcp - would be more quiet but    */
/*              also more fragile                                           */

int test_port(unsigned short p)
{
  int fd;
  struct sockaddr_in addr;
  int addrlen;

  addr.sin_family = AF_INET;
  addr.sin_port = htons(p);
  addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  addrlen = sizeof(addr);

  fd = socket(AF_INET, SOCK_STREAM, 0);

  if (fd < 0) {
    return -1;
  }
  if (connect(fd, (struct sockaddr *) &(addr), addrlen)) {
    close(fd);
    if (errno == ECONNREFUSED) {
      return 0;
    } else {
      return -1;
    }
  } else {			/* local port is available */
    close(fd);
    return 1;
  }

}

/****************************************************************************/
/* Does       : system() replacement without involving a shell              */
/* Parameters : s - string to be executed                                   */

int strexec(char *s)
{
  int i, j, e;
  char **a;

  i = 0;
  j = 1;
  e = 0;
  while (s[i] != '\0') {
    if (isspace(s[i])) {
      j = 1;
    } else {
      if (j) {
	e++;
	j = 0;
      }
    }
    i++;
  }
#ifdef TRACE
  fprintf(stderr, "strexec(): <%s> has %d elements\n", s, e);
#endif
  e++;

  a = malloc(sizeof(char *) * e);
  if (a) {
    i = 0;
    j = 1;
    e = 0;
    while (s[i] != '\0') {
      if (isspace(s[i])) {
	s[i] = '\0';
	j = 1;
      } else {
	if (j) {
	  a[e] = s + i;
	  e++;
	  j = 0;
	}
      }
      i++;
    }
    a[e] = NULL;
    execvp(s, a);
#ifdef TRACE
    fprintf(stderr, "strexec(): execvp(%s,...) failed: %s\n", s,
	    strerror(errno));
#endif
  }
  return -1;
}

void drop_parent(char *name, char *id)
{
  uid_t uid = 0;
  gid_t gid = 0;
  struct passwd *pw;

  if (id) {
    uid = atoi(id);
    if (uid <= 0) {
      pw = getpwnam(id);
      if (pw == NULL) {
	fprintf(stderr, "%s: unable to look up user %s\n", name, id);
	exit(1);
      } else {
	uid = pw->pw_uid;
	gid = pw->pw_gid;
	/* docs are unclear - do I need to free pw ? */
	/* free(pw); */
      }
    } else {
      gid = uid;
    }
    if (setgid(gid) || setuid(uid)) {
      fprintf(stderr, "%s: unable to change id to %d/%d: %s\n", name, uid,
	      gid, strerror(errno));
      exit(1);
    }
  }
}

void drop_fork(char *name)
{
  struct rlimit r;

  r.rlim_cur = 0;
  r.rlim_max = 0;

  if (setrlimit(RLIMIT_NPROC, &r)) {
    fprintf(stderr, "%s: unable to reduce process limit: %s\n", name,
	    strerror(errno));
    exit(1);
  }
}

void drop_root(char *name, char *id, char *rootdir)
{
  uid_t uid = 0;
  gid_t gid = 0;
  struct passwd *pw;

  if (id) {
    uid = atoi(id);
    if (uid <= 0) {
      pw = getpwnam(id);
      if (pw == NULL) {
	fprintf(stderr, "%s: unable to look up user %s\n", name, id);
	exit(1);
      } else {
	uid = pw->pw_uid;
	gid = pw->pw_gid;
	/* docs are unclear - do I need to free pw ? */
	/* free(pw); */
      }
    } else {
      gid = uid;
    }
  } else {
    if (getuid() == 0) {
      fprintf(stderr, "%s: consider having %s drop root privileges\n",
	      name, name);
    }
  }

  if (rootdir != NULL) {	/* do chroot */
    if (chroot(rootdir)) {
      fprintf(stderr, "%s: unable to change root to %s: %s\n", name,
	      rootdir, strerror(errno));
      exit(1);
    }
  }
  chdir("/");

  if (id) {			/* now change id */
    if (setgid(gid) || setuid(uid)) {
      fprintf(stderr, "%s: unable to change id to %d/%d: %s\n", name, uid,
	      gid, strerror(errno));
      exit(1);
    }
  }
}
