#include <stdlib.h>
#include <stdio.h>

#include <idsa_internal.h>

#include "types.h"

#include "table.h"
#include "branch.h"
#include "set.h"
#include "queue.h"
#include "mod.h"

UNION_BRANCH *branch_new(RULE_BODY * r, int t)
{
  UNION_BRANCH *result;

  result = malloc(sizeof(UNION_BRANCH));
  if (result) {
    result->b_compares = 0;
    result->b_runs = 0;
    result->b_type = t;
    result->b_first = r;
    result->b_last = r;
    result->b_prev = NULL;
    result->b_next = NULL;
    switch (result->b_type) {
    case BRANCH_TYPE_TABLE:
      result->b_un.b_table = NULL;
      break;
    case BRANCH_TYPE_QUEUE:
      result->b_un.b_queue = NULL;
      break;
    case BRANCH_TYPE_MOD:
      result->b_un.b_mod = NULL;
      break;
    default:
#ifdef TRACE
      fprintf(stderr, "branch_new(): unknown type %d\n", result->b_type);
#endif
      free(result);
      result = NULL;
      break;
    }
  }
  return result;
}

void branch_free(UNION_BRANCH * b)
{
  UNION_BRANCH *c;

  c = b;
  while (c) {
    b = c;
    c = c->b_next;
    switch (b->b_type) {
    case BRANCH_TYPE_TABLE:
      table_free(b->b_un.b_table);
      b->b_un.b_table = NULL;
      break;
    case BRANCH_TYPE_QUEUE:
      queue_match_free(b->b_un.b_queue);
      b->b_un.b_queue = NULL;
      break;
    case BRANCH_TYPE_MOD:
      mod_free(b->b_un.b_mod);
      b->b_un.b_mod = NULL;
      break;
    }
    b->b_next = NULL;
    b->b_prev = NULL;

    b->b_first = NULL;
    b->b_last = NULL;

    free(b);
  }
}

/****************************************************************************/
/* Does       : exchanges current branch with previous if they are used     */
/*              for the first time in the same rule. An attempt at          */
/*              dynamic optimization                                        */
/* Returns    : b if unchanged, otherwise a (since a->b_next is next target) */

UNION_BRANCH *branch_reorder(STATE_SET * s, UNION_BRANCH * b)
{
  UNION_BRANCH *a;

  a = b;

#ifdef TRACE
  fprintf(stderr, "branch_reorder(): before\n");
  branch_dump(s, stderr);
#endif

  /* condition to be met for exchange, including cost comparisons */
  if (b && b->b_prev && (b->b_prev->b_first == b->b_first)
      && (b->b_prev->b_type >= b->b_type)) {
    a = b->b_prev;

    b->b_prev = a->b_prev;
    if (a->b_prev) {
      a->b_prev->b_next = b;
    } else {
      s->s_branch = b;
    }

    a->b_next = b->b_next;
    if (b->b_next) {
      b->b_next->b_prev = a;
    }
    a->b_prev = b;
    b->b_next = a;
  }
#ifdef TRACE
  fprintf(stderr, "branch_reorder(): after\n");
  branch_dump(s, stderr);
#endif

  return a;
}

void branch_append(STATE_SET * s, UNION_BRANCH * b)
{
  UNION_BRANCH *c;

  b->b_next = NULL;

  if (s->s_branch) {
    c = s->s_branch;
    while (c->b_next) {
      c = c->b_next;
    }
    c->b_next = b;
    b->b_prev = c;
  } else {
    b->b_prev = NULL;
    s->s_branch = b;
  }
}

int branch_trigger(UNION_BRANCH * b, IDSA_EVENT * e)
{
  int result = 0;

#ifdef TRACE
  fprintf(stderr,
	  "branch_trigger(): triggering type <0x%02x>, ptr=<%p> [%d-%d]\n",
	  b->b_type, b, b->b_first->r_number, b->b_last->r_number);
#endif
  switch (b->b_type) {
  case BRANCH_TYPE_TABLE:
    result = table_trigger(b->b_un.b_table, e);
    break;
  case BRANCH_TYPE_QUEUE:
    result = queue_trigger(b->b_un.b_queue, e);
    break;
  case BRANCH_TYPE_MOD:
    result = mod_trigger(b->b_un.b_mod, e);
    break;
  }

  b->b_compares += result;
  b->b_runs++;

#ifdef TRACE
  fprintf(stderr, "branch_append(): compares=%d, runs=%d\n", b->b_compares,
	  b->b_runs);
#endif

  return result;
}

void branch_dump(STATE_SET * s, FILE * fp)
{
  UNION_BRANCH *b;

  b = s->s_branch;

  fprintf(fp, "branch_dump(): starting <%p>\n", b);

  while (b) {
    fprintf(fp, "<%p] [%p] [%p>: %p[%d] %p[%d]\n", b->b_prev, b, b->b_next,
	    b->b_first, (b->b_first) ? (b->b_first->r_number) : 0,
	    b->b_last, (b->b_last) ? (b->b_last->r_number) : 0);
    b = b->b_next;
  }

  fprintf(fp, "branch_dump(): done\n");
}
