
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <idsa_internal.h>

#include "rule.h"
#include "queue.h"
#include "parse.h"
#include "branch.h"
#include "set.h"

static QUEUE_MATCH *queue_match_new(char *n, QUEUE_VAR * qv);
static QUEUE_VAR *queue_var_new(char *n, unsigned int t);
static QUEUE_ITEM *queue_item_new(char *n, unsigned int t);
static QUEUE_KEEP *queue_keep_new(char *n, QUEUE_VAR * qv);

static QUEUE_VAR *queue_var_find(STATE_SET * s, char *v, unsigned int t);

static void queue_add(QUEUE_VAR * qv, IDSA_UNIT * u, time_t t);
static void queue_expire(QUEUE_VAR * qv);
static void queue_delete(QUEUE_VAR * qv, QUEUE_ITEM * qi);

/****************************************************************************/

void queue_item_free(QUEUE_ITEM * qi)
{
  if (qi->i_unit) {
    idsa_unit_free(qi->i_unit);
    qi->i_unit = NULL;
  }
  free(qi);
}

void queue_keep_free(QUEUE_KEEP * qk)
{
  qk->k_queue = NULL;
  free(qk);
}

void queue_match_free(QUEUE_MATCH * qm)
{
  if (qm->m_match) {
    free(qm->m_match);
    qm->m_match = NULL;
  }
  qm->m_count = 0;
  qm->m_queue = NULL;

  free(qm);
}

/****************************************************************************/
/* Does       : delete v and all v->v_next                                  */

void queue_var_free(QUEUE_VAR * qv)
{
  QUEUE_ITEM *qi, *qj;
  QUEUE_VAR *qw;

  qw = qv;
  while (qw) {
    qv = qw;
    qw = qw->v_next;

    qi = qv->v_orderhead;
    while (qi) {
      qj = qi;
      qi = qi->i_ordernext;
      queue_item_free(qj);
    }
    qv->v_orderhead = NULL;
    qv->v_ordertail = NULL;

    qi = qv->v_free;
    while (qi) {
      qj = qi;
      qi = qi->i_ordernext;
      queue_item_free(qj);
    }
    qv->v_free = NULL;

    qv->v_timehead = NULL;
    qv->v_timetail = NULL;

    qv->v_next = NULL;

    free(qv);
  }
}

RULE_MATCH *queue_match_make(STATE_SET * s, char *v, unsigned int t,
			     char *n)
{
  UNION_BRANCH *b;
  QUEUE_VAR *qv;
  QUEUE_MATCH *qm;
  RULE_MATCH *m, **p;

  b = s->s_branch;
  qv = NULL;
  qm = NULL;
  m = NULL;

  while (b) {
    if (b->b_type == BRANCH_TYPE_QUEUE) {
      qm = b->b_un.b_queue;

      if ((strncmp(qm->m_name, v, IDSA_M_NAME) == 0)
	  && (strncmp(qm->m_queue->v_var, v, IDSA_M_NAME) == 0)) {
#ifdef TRACE
	fprintf(stderr, "queue_make_match(): found matching name <%s>\n",
		v);
#endif
	b->b_last = s->s_currentrule;
	b = NULL;		/* bomb - a break in the wrong place blew up ariane */
      } else {
	b = b->b_next;
	qm = NULL;
      }
    } else {
      b = b->b_next;
    }
  }

  if (qm) {
    if ((qm->m_queue->v_type != t) && (t != IDSA_T_NULL)) {
      parse_error(s,
		  "consistency error: conflicting types for variable %s",
		  v);
      qm = NULL;
    } else {
      t = qm->m_queue->v_type;
    }
  } else {
    qv = queue_var_find(s, v, t);
    if (qv) {
      qm = queue_match_new(n, qv);
      if (qm) {
	b = branch_new(s->s_currentrule, BRANCH_TYPE_QUEUE);
	if (b) {
	  b->b_un.b_queue = qm;
	  branch_append(s, b);
	} else {
	  parse_errormem(s, sizeof(UNION_BRANCH));
	  queue_match_free(qm);
	  qm = NULL;
	}
      } else {
	parse_errormem(s, sizeof(QUEUE_MATCH));
      }
    }
  }

  if (qm) {
    m = match_new(RULE_GATE_NONE);
    if (m) {
      p = realloc(qm->m_match, sizeof(RULE_MATCH *) * (qm->m_count + 1));
      if (p) {
	p[qm->m_count] = m;
	qm->m_count++;
	qm->m_match = p;
      } else {
	parse_errormem(s, sizeof(RULE_MATCH *) * (qm->m_count + 1));
	match_free(m);
	m = NULL;
      }
    } else {
      parse_errormem(s, sizeof(RULE_MATCH));
    }
  }				/* else some failure */
  return m;
}

void queue_set_size(STATE_SET * s, QUEUE_KEEP * qk, int z)
{
  QUEUE_VAR *qv;
  QUEUE_ITEM *qi;

  qv = qk->k_queue;

  while (qv->v_size < z) {
    qi = queue_item_new(qk->k_name, qv->v_type);
    if (qi) {
      qi->i_ordernext = qv->v_free;
      qv->v_free = qi;
      qv->v_size++;
    } else {
      z = qv->v_size;
      parse_errormem(s, sizeof(QUEUE_ITEM));
    }
  }
}

void queue_set_timeout(STATE_SET * s, QUEUE_KEEP * qk, int t)
{
  qk->k_timeout = t;
}

QUEUE_KEEP *queue_keep_make(STATE_SET * s, char *v, unsigned int t,
			    char *n)
{
  QUEUE_KEEP *qk = NULL;
  QUEUE_VAR *qv;

#ifdef TRACE
  fprintf(stderr, "queue_keep_make(): keeping <%s:%d> in variable <%s>\n",
	  n, t, v);
#endif

  qv = queue_var_find(s, v, t);
  if (qv) {

    qk = queue_keep_new(n, qv);
    if (qk == NULL) {
      parse_errormem(s, sizeof(QUEUE_KEEP));
    }
  }				/* else some failure and s_error set */
  return qk;
}

/****************************************************************************/
/** Run time stuff starts here **********************************************/
/****************************************************************************/

/****************************************************************************/
/* Does       : sets list of matches to true or false                       */
/* Returns    : number of comparisons made                                  */

int queue_trigger(QUEUE_MATCH * qm, IDSA_EVENT * e)
{
  QUEUE_ITEM *qi;
  QUEUE_VAR *qv;
  IDSA_UNIT *u;
  int r, i;
  int result = 0;
  int found = 0;
#ifdef TRACE
  char buffer[IDSA_M_LONG];
  int l;
#endif

  if (qm->m_reserved < idsa_required_count()) {
    u = idsa_event_unitbynumber(e, qm->m_reserved);
  } else {
    u = idsa_event_unitbyname(e, qm->m_name);
  }

  qv = qm->m_queue;
  queue_expire(qv);

#ifdef TRACE
  fprintf(stderr, "queue_trigger(): looking at $%s\n", qm->m_queue->v_var);
#endif

  if (u && (qv->v_type == idsa_unit_type(u))) {
    qi = qv->v_orderhead;
    while (qi != NULL) {
      result++;
      r = idsa_unit_compare(u, qi->i_unit);
#ifdef TRACE
      fprintf(stderr, "queue_trigger(): comparision of ");
      l = idsa_unit_print(u, buffer, IDSA_M_LONG, 0);
      buffer[l] = '\0';
      fprintf(stderr, buffer);
      fprintf(stderr, " and ");
      l = idsa_unit_print(qi->i_unit, buffer, IDSA_M_LONG, 0);
      buffer[l] = '\0';
      fprintf(stderr, buffer);
      fprintf(stderr, " yields %d\n", r);
#endif
      switch (r) {
      case IDSA_COMPARE_EQUAL:
      case IDSA_COMPARE_INTERSECT:
	if (qi->i_orderprev) {	/* move successful match to front in hope of matching again soon */
	  /* snip out of old position */
	  qi->i_orderprev->i_ordernext = qi->i_ordernext;
	  if (qi->i_ordernext) {
	    qi->i_ordernext->i_orderprev = qi->i_orderprev;
	  } else {
	    qv->v_ordertail = qi->i_orderprev;
	  }

	  /* insert at front, certain that there is a front */
	  qv->v_orderhead->i_orderprev = qi;
	  qi->i_ordernext = qv->v_orderhead;
	  qv->v_orderhead = qi;
	  qi->i_orderprev = NULL;
	}
	qi = NULL;		/* bomb out of loop */
	found = 1;
	break;
      default:
	qi = qi->i_ordernext;
	break;
      }
    }
  }
  if (found) {
    for (i = 0; i < qm->m_count; i++) {
      match_fire_true(qm->m_match[i]);
    }
  } else {
    for (i = 0; i < qm->m_count; i++) {
      match_fire_false(qm->m_match[i]);
    }
  }

  return result;
}

int queue_run(QUEUE_KEEP * qk, IDSA_EVENT * e)
{
  int result = 0;
  QUEUE_VAR *qv;
  IDSA_UNIT *u;

  qv = qk->k_queue;

#ifdef TRACE
  fprintf(stderr, "queue_keep_run(): dump before insert\n");
  queue_dump(qv, stderr);
  idsa_event_dump(e, stderr);
#endif

  if (qk->k_reserved < idsa_required_count()) {
    u = idsa_event_unitbynumber(e, qk->k_reserved);
  } else {
    u = idsa_event_unitbyname(e, qk->k_name);
  }

  if (u && (qv->v_type == idsa_unit_type(u))) {	/* have source, now append */
    queue_expire(qv);		/* less is always better */
    queue_add(qv, u, qk->k_timeout);
#ifdef TRACE
    fprintf(stderr, "queue_keep_run(): dump after insert\n");
    queue_dump(qv, stderr);
#endif
  } else {
#ifdef TRACE
    fprintf(stderr, "queue_keep_run(): type mismatch\n");
#endif
    result++;
  }

  return result;
}

/****************************************************************************/
/* Does       : display the queue variable (and its items)                  */

void queue_dump(QUEUE_VAR * qv, FILE * f)
{
  QUEUE_ITEM *qi, *qj;
  char buffer[IDSA_M_LONG];
  int l;

  fprintf(f, "queue_dump(): type=<%d>, variable=<%s>\n", qv->v_type,
	  qv->v_var);
  fprintf(f, "queue_dump(): order head=%p, tail=%p\n", qv->v_orderhead,
	  qv->v_ordertail);

  qi = qv->v_orderhead;
  qj = NULL;
  while (qi) {
    l = idsa_unit_print(qi->i_unit, buffer, IDSA_M_LONG, 0);
    if (l > 0) {
      buffer[l] = '\0';
      fprintf(f, "queue_dump(): [%p] %s:%s [%p]", qi->i_orderprev,
	      idsa_unit_name(qi->i_unit), buffer, qi->i_ordernext);
    }
    if (qj != qi->i_orderprev) {
      fprintf(f, "[%p!=%p]", qj, qi->i_orderprev);
    }
    qj = qi;
    qi = qi->i_ordernext;
    fputc('\n', f);
  }

  fprintf(f, "queue_dump(): time head=%p, tail=%p\n", qv->v_timehead,
	  qv->v_timetail);
  qi = qv->v_timehead;
  qj = NULL;
  while (qi) {
    l = idsa_unit_print(qi->i_unit, buffer, IDSA_M_LONG, 0);
    if (l > 0) {
      buffer[l] = '\0';
      fprintf(f, "queue_dump(): [%p] %s:%d", qi, buffer,
	      (int) (qi->i_timeout));
    }
    if (qj != qi->i_timeprev) {
      fprintf(f, "[%p!=%p]", qj, qi->i_orderprev);
    }
    qj = qi;
    qi = qi->i_timenext;
    fputc('\n', f);
  }

}

/****************************************************************************/

static QUEUE_KEEP *queue_keep_new(char *n, QUEUE_VAR * qv)
{
  QUEUE_KEEP *qk;
  qk = malloc(sizeof(QUEUE_KEEP));
  if (qk) {
    strncpy(qk->k_name, n, IDSA_M_NAME - 1);
    qk->k_name[IDSA_M_NAME - 1] = '\0';

    qk->k_reserved = idsa_reserved_numberbyname(n);
    qk->k_queue = qv;
    qk->k_timeout = 0;
  }
  return qk;
}

static QUEUE_ITEM *queue_item_new(char *n, unsigned int t)
{
  QUEUE_ITEM *qi;
  qi = malloc(sizeof(QUEUE_ITEM));
  if (qi) {
    qi->i_timenext = NULL;
    qi->i_timeprev = NULL;

    qi->i_ordernext = NULL;
    qi->i_orderprev = NULL;

    qi->i_unit = idsa_unit_new(n, t, NULL);

    if (qi->i_unit == NULL) {
      free(qi);
      qi = NULL;
    }
  }
  return qi;
}

static QUEUE_MATCH *queue_match_new(char *n, QUEUE_VAR * qv)
{
  QUEUE_MATCH *qm;
  qm = malloc(sizeof(QUEUE_MATCH));
  if (qm) {
    strncpy(qm->m_name, n, IDSA_M_NAME - 1);
    qm->m_name[IDSA_M_NAME - 1] = '\0';

    qm->m_reserved = idsa_reserved_numberbyname(n);

    qm->m_queue = qv;

    qm->m_count = 0;
    qm->m_match = NULL;
  }
  return qm;
}

static QUEUE_VAR *queue_var_new(char *n, unsigned int t)
{
  QUEUE_VAR *qv;
  qv = malloc(sizeof(QUEUE_VAR));
  if (qv) {
    qv->v_type = t;

    strncpy(qv->v_var, n, IDSA_M_NAME - 1);
    qv->v_var[IDSA_M_NAME - 1] = '\0';

    qv->v_size = 0;

    qv->v_orderhead = NULL;
    qv->v_ordertail = NULL;
    qv->v_timehead = NULL;
    qv->v_timetail = NULL;
    qv->v_free = NULL;

    qv->v_next = NULL;
  }
  return qv;
}

/****************************************************************************/
/* Does       : find or create a variable                                   */

static QUEUE_VAR *queue_var_find(STATE_SET * s, char *v, unsigned int t)
{
  QUEUE_VAR *qv;

  qv = s->s_queue;
  while ((qv != NULL) && (strncmp(qv->v_var, v, IDSA_M_NAME) != 0)) {
    qv = qv->v_next;
  }

  if (qv) {
    if ((qv->v_type != t) && (t != IDSA_T_NULL)) {
      parse_error(s,
		  "consistency error: conflicting types for variable %s",
		  v);
      qv = NULL;
    } else {
      t = qv->v_type;
    }
  } else {
    if (t != IDSA_T_NULL) {
      qv = queue_var_new(v, t);
      if (qv) {
	qv->v_next = s->s_queue;
	s->s_queue = qv;
      } else {
	parse_errormem(s, sizeof(QUEUE_VAR));
      }
    } else {
      parse_error(s, "require a type for $%s", v);
    }
  }

  return qv;
}

static void queue_add(QUEUE_VAR * qv, IDSA_UNIT * u, time_t t)
{
  QUEUE_ITEM *qi, *qj;
  int r;

  /* FIXME: might need better, but more expensive duplicate check */
  qi = NULL;
  if (qv->v_orderhead) {	/* check that it has not been added recently */
    r = idsa_unit_compare(u, qv->v_orderhead->i_unit);
    switch (r) {
    case IDSA_COMPARE_EQUAL:
    case IDSA_COMPARE_INTERSECT:
      qi = qv->v_orderhead;

      /* chop out of time list, since timer may need to be reset */
      if (qi->i_timenext) {	/* fix up next one */
	qi->i_timenext->i_timeprev = qi->i_timeprev;
      } else {
	qv->v_timetail = qi->i_timeprev;
      }
      if (qi->i_timeprev) {	/* fix up previous one */
	qi->i_timeprev->i_timenext = qi->i_timenext;
      } else {			/* no previous, so we must be first, better fix head */
	qv->v_timehead = qi->i_timenext;
      }
      break;
    }
  }

  if (qi == NULL) {		/* no hit, better insert it properly */

    if (qv->v_free == NULL) {	/* no space, turf item closest to expire */
      /* FIXME: good place to check for duplicates */
      queue_delete(qv, qv->v_timehead);
    }
    /* delete is not supposed to fail, so we should have space */
    qi = qv->v_free;
    qv->v_free = qi->i_ordernext;

    idsa_unit_copy(qi->i_unit, u);

#ifdef TRACE
    fprintf(stderr, "queue_add(): copied unit %p=<%s:%d> to %p=<%s:%d>\n",
	    u, idsa_unit_name(u), idsa_unit_type(u), qi,
	    idsa_unit_name(qi->i_unit), idsa_unit_type(qi->i_unit));
#endif

    /* make it order head */
    qi->i_orderprev = NULL;
    if (qv->v_orderhead) {
      qv->v_orderhead->i_orderprev = qi;
    } else {			/* tail also has to point to NULL, so update */
      qv->v_ordertail = qi;
    }
    qi->i_ordernext = qv->v_orderhead;
    qv->v_orderhead = qi;
  }

  /* now the order is connected up, but time is not */

  if (t) {
    qi->i_timeout = time(NULL) + t;
  } else {
    /* WARNING: is this really the highest time possible ? */
    qi->i_timeout = LONG_MAX;
  }

  /* search time tail from back */
  qj = qv->v_timetail;
  while ((qj != NULL) && (qj->i_timeout > qi->i_timeout)) {
    qj = qj->i_timeprev;
  }
  if (qj) {			/* insert after qj - so not at start */
    qi->i_timeprev = qj;
    qi->i_timenext = qj->i_timenext;
    if (qi->i_timenext) {
      qi->i_timenext->i_timeprev = qi;
    } else {			/* we are last, better update tail */
      qv->v_timetail = qi;
    }
    qj->i_timenext = qi;
  } else {			/* wow, right at start */
    qi->i_timeprev = NULL;
    if (qv->v_timehead) {
      qv->v_timehead->i_timeprev = qi;
    } else {
      qv->v_timetail = qi;
    }
    qi->i_timenext = qv->v_timehead;
    qv->v_timehead = qi;
  }
}

/****************************************************************************/
/* Does       : removes those items whose timer has expired                 */

static void queue_expire(QUEUE_VAR * qv)
{
  time_t t;

  t = time(NULL);

  while ((qv->v_timehead != NULL) && (qv->v_timehead->i_timeout < t)) {
#ifdef TRACE
    fprintf(stderr, "queue_expire(): expiring %p\n", qv->v_timehead);
#endif
    queue_delete(qv, qv->v_timehead);
  }
}

/****************************************************************************/
/* Does       : remove element from v_time and v_order and drop it into     */
/*              v_free                                                      */

static void queue_delete(QUEUE_VAR * qv, QUEUE_ITEM * qi)
{
  if (qi->i_ordernext) {
    qi->i_ordernext->i_orderprev = qi->i_orderprev;
  } else {
    qv->v_ordertail = qi->i_orderprev;
  }
  if (qi->i_orderprev) {
    qi->i_orderprev->i_ordernext = qi->i_ordernext;
  } else {
    qv->v_orderhead = qi->i_ordernext;
  }

  if (qi->i_timenext) {		/* fix up next one */
    qi->i_timenext->i_timeprev = qi->i_timeprev;
  } else {
    qv->v_timetail = qi->i_timeprev;
  }
  if (qi->i_timeprev) {		/* fix up previous one */
    qi->i_timeprev->i_timenext = qi->i_timenext;
  } else {			/* no previous, so we must be first, better fix head */
    qv->v_timehead = qi->i_timenext;
  }

  /* link into singly linked list of free elements */
  qi->i_ordernext = qv->v_free;
  qi->i_orderprev = NULL;

  qi->i_timenext = NULL;
  qi->i_timeprev = NULL;

  qv->v_free = qi;
}
